from abc import abstractmethod
from asyncio import iscoroutinefunction
from typing import ClassVar, Optional, Callable, Coroutine, Any, Union

from pydantic import BaseModel, ValidationError, Field

from autopack.utils import run_args_from_args_schema, acall_llm, call_llm


class Pack(BaseModel):
    arbitrary_types_allowed = True

    class Config:
        arbitrary_types_allowed = True

    ## Required

    # The name of the tool that will be provided to the LLM
    name: ClassVar[str]
    # The description of the tool which is passed to the LLM
    description: ClassVar[str]

    ## Optional

    # Any pip packages this Pack depends on.
    dependencies: ClassVar[Optional[list[str]]] = None
    # A list of Pack IDs needed for this tool to function effectively (e.g. `write_file` depends on `read_file`)
    depends_on: ClassVar[Optional[list[str]]] = None
    # Enhances tool selection by grouping this tool with other tools of the same category
    categories: ClassVar[Optional[list[str]]] = None
    # If this tool has side effects that cannot be undone (e.g. sending an email)
    reversible: ClassVar[bool] = True
    # A Pydantic BaseModel describing the Pack's run arguments
    args_schema: ClassVar[Optional[type[BaseModel]]] = None

    llm: Optional[Callable[[str], str]] = Field(
        None, description="A callable function to call an LLM (string in string out)"
    )
    allm: Union[None, Callable[[str], str], Coroutine[Any, Any, str]] = Field(
        None, description="An asynchronous callable function to call an LLM (string in string out)"
    )

    def __init__(self, **data):
        super().__init__(**data)
        if self.llm and not callable(self.llm):
            raise TypeError(f"LLM object {self.llm} must be callable")

        if self.allm and not iscoroutinefunction(self.allm):
            raise TypeError(f"Async LLM object {self.llm} must be async and callable")

        if self.name is None:
            raise TypeError(f"Class {self.__class__.__name__} must define 'name' as a class variable")
        if self.description is None:
            raise TypeError(f"Class {self.__class__.__name__} must define 'description' as a class variable")

    def run(self, *args, **kwargs) -> str:
        """Execute the _run function of the subclass, verifying the arguments. (Will eventually do callbacks or some
        such)

        Args: **kwargs (dict): The arguments to pass to _run. Each key should be the name of an argument,
        and the value should be the value of the argument.

        Returns: The response from the _run function of the subclass
        """
        try:
            # Validate the arguments
            self.validate_tool_args(**kwargs)
        except ValidationError as e:
            # If a ValidationError is raised, the arguments are invalid
            error_list = ". ".join([f"{err['loc'][0]}: {err['msg']}" for err in e.errors()])
            return f"Error: Invalid arguments. Details: {error_list}"

        return self._run(*args, **kwargs)

    async def arun(self, *args, **kwargs):
        """Asynchronously execute the _arun function of the subclass, verifying the arguments. (Will eventually do
        callbacks or some such)

        Args:
            **kwargs (dict): The arguments to pass to _arun. Each key should be the name of an argument,
        and the value should be the value of the argument.

        Returns:
            str: The response from the _arun function of the subclass
        """
        try:
            # Validate the arguments
            self.validate_tool_args(**kwargs)
        except ValidationError as e:
            # If a ValidationError is raised, the arguments are invalid
            error_list = ". ".join([f"{err['loc'][0]}: {err['msg']}" for err in e.errors()])
            return f"Error: Invalid arguments. Details: {error_list}"

        return await self._arun(*args, **kwargs)

    @abstractmethod
    def _run(self, *args, **kwargs):
        pass

    @abstractmethod
    def _arun(self, *args, **kwargs):
        pass

    @property
    def args(self) -> dict:
        """Turn the args schema into a dict that's easier to work with"""
        if not self.args_schema:
            return {}
        return run_args_from_args_schema(self.args_schema)

    def call_llm(self, prompt: str) -> str:
        if self.llm is None:
            return "No LLM available, cannot proceed"
        return call_llm(prompt, self.llm)

    async def acall_llm(self, prompt: str) -> str:
        if self.allm is None:
            return self.call_llm(prompt)
        return await acall_llm(prompt, self.allm)

    def validate_tool_args(self, **kwargs):
        """Validate the arguments against the args_schema model.

        Args:
            kwargs (dict[str, Any]): The arguments to validate.
        Returns:
            True if the arguments are valid.
        Raises:
            ValidationError If any arguments are invalid.
        """
        self.args_schema(**kwargs)

        return True



import json

from autopack.api import pack_search


def print_search(query: str):
    matching_packs = pack_search(query)
    for pack in matching_packs:
        print("--------")
        print(f"Pack ID:      {pack.pack_id}")
        print(f"Dependencies: {', '.join(pack.dependencies)}")
        print(f"Description:  {pack.name}")
        print(f"Run Args:     {json.dumps(pack.run_args)}")



from typing import Union

from autopack.api import get_pack_details
from autopack.errors import AutoPackError, AutoPackNotFoundError
from autopack.pack import Pack
from autopack.utils import fetch_pack_object, load_metadata_file


def try_get_pack(pack_id: str, remote=False) -> Union[type[Pack], None]:
    """
    Get a pack based on its ID. Same as `get_pack` but does not raise an Exception. If there is a problem finding or
    loading a pack it will return None.

    Args:
        pack_id (str): The ID of the pack to fetch.
        quiet (bool, Optional): If True, won't print any output
        remote (bool, Optional): If True, will make network requests to fetch pack metadata

    Returns:
        Pack or None: The fetched pack, None if the pack could not be loaded
    """

    try:
        return get_pack(pack_id, remote=remote)
    except AutoPackError:
        return None


def get_all_installed_packs():
    """Returns all of the"""
    metadata = load_metadata_file()
    pack_ids = list(metadata.keys())
    return try_get_packs(pack_ids, remote=False)


def try_get_packs(pack_ids: list[str], remote=False) -> list[type[Pack]]:
    """
    Get a list of packs based on their IDs

    Args:
        pack_ids (list[str]): The IDs of the packs to fetch.
        quiet (bool, Optional): If True, won't print any output
        remote (bool, Optional): If True, will make network requests to fetch pack metadata

    Returns:
        list[Pack]: The successfully fetched packs
    """
    packs = []
    for pack_id in pack_ids:
        pack = try_get_pack(pack_id, remote)
        if pack:
            packs.append(pack)

    return packs


def get_pack(pack_id: str, remote=False) -> type[Pack]:
    """
    Get a pack based on its ID.

    Args:
        pack_id (str): The ID of the pack to fetch.
        remote (bool, Optional): If True, will make network requests to fetch pack metadata

    Returns:
        Pack: The fetched pack

    Raises:
        AutoPackFetchError: If there was an error during the data fetch.
        AutoPackNotInstalledError: If the pack was found but not installed.
        AutoPackNotFoundError: If no pack matching that ID was found.
        AutoPackLoadError: If the pack was found but there was an error importing or finding the pack class.
    """
    pack_data = get_pack_details(pack_id, remote=remote)

    if not pack_data:
        raise AutoPackNotFoundError()

    return fetch_pack_object(pack_data)



import re
from typing import Callable, Union, Optional

from langchain.chat_models.base import BaseChatModel

from autopack import Pack
from autopack.get_pack import get_all_installed_packs
from autopack.prompts import GET_MORE_TOOLS_TEMPLATE, TOOL_SELECTION_TEMPLATE
from autopack.utils import functions_bulleted_list, call_llm


def select_packs_prompt(task_description: str, function_request: Optional[str] = None) -> str:
    """
    Generate a prompt for the pack selection process based on the task description and an optional function request.

    Args:
        task_description (str): A description of the task to be used when selecting tools.
        function_request (Optional[str]): A specific type of function asked for (e.g. a `get_more_tools` function).

    Returns:
        str: A prompt that can be fed to the LLM for pack selection.
    """

    installed_packs = get_all_installed_packs()

    if function_request:
        return TOOL_SELECTION_TEMPLATE.format(task=task_description, functions=functions_bulleted_list(installed_packs))

    return GET_MORE_TOOLS_TEMPLATE.format(
        task=task_description,
        functions=functions_bulleted_list(installed_packs),
        functions_request=function_request,
    )


def select_packs(
    task_description: str,
    llm: Union[BaseChatModel, Callable],
    function_request: Optional[str] = None,
) -> list[type[Pack]]:
    """Given a user input describing the task they wish to accomplish, return a list of Pack IDs that the given LLM
    thinks will be suitable for this task.

    This is good for a "pre-processing" step after receiving the task but before trying to solve it. This allows you
    to benefit from the wide tool selection while keeping your token usage low.

    You can then further filter, install the packs if desired, and then fetch them using get_pack().

    # TODO: Include user-provided packs into selection

    Args:
        task_description (str): A description of the task to be used when selecting tools
        llm (BaseChatModel): An LLM which will be used to evaluate the selection
        function_request (Optional[str]): A specific type of function asked for (e.g. a `get_more_tools` function)

    Returns:
        list[str]: A list of selected Pack IDs
    """
    prompt = select_packs_prompt(task_description, function_request)

    response = call_llm(prompt, llm)

    return parse_selection_response(response)


def parse_selection_response(response: str) -> list[type[Pack]]:
    """
    Parse the response from the LLM and extract pack IDs.

    The response is split by commas and newlines, and any arguments provided in the response are removed.

    Args:
        response (str): The response from the LLM.

    Returns:
        list[str]: A list of parsed pack IDs.
    """
    pack_names = [r.split("(")[0].strip() for r in re.split(r"(?<=\w),|\n", response)]
    installed_packs = get_all_installed_packs()
    selected_packs = []
    for pack_name in pack_names:
        try:
            selected_pack = next(pack for pack in installed_packs if pack.name == pack_name)
            selected_packs.append(selected_pack)
        except StopIteration:
            # This means that the pack selected is not installed. This error should've been caught elsewhere
            continue

    return selected_packs



class AutoPackError(Exception):
    pass


class AutoPackFetchError(AutoPackError):
    pass


class AutoPackNotFoundError(AutoPackError):
    pass


class AutoPackNotInstalledError(AutoPackError):
    pass


class AutoPackLoadError(AutoPackError):
    pass


class AutoPackInstallationError(AutoPackError):
    pass
